


//////////////////////////////////////////////////////////////////
//                                                              //
//           PLINK (c) 2005-2007 Shaun Purcell                  //
//                                                              //
// This file is distributed under the GNU General Public        //
// License, Version 2.  Please see the file COPYING for more    //
// details                                                      //
//                                                              //
//////////////////////////////////////////////////////////////////


#include <iostream>
#include <iomanip>
#include <string>
#include <set>
#include <cmath>
#include "options.h"
#include "helper.h"
#include "plink.h"
#include "phase.h"


//////////////////////////////////////////
// Helper class to store and sort proxies

class ProxyResult {
public:

  string name;
  double f;
  double r2;
  double odds;
  double chisq;
  double pvalue;

  ProxyResult(string n, double frq, double r, double o, double c, double p)
    : name(n), f(frq), r2(r), odds(o), chisq(c), pvalue(p) { } 

  bool operator< (const ProxyResult & b) const
    {
      return ( pvalue == b.pvalue ? 
	       name < b.name : 
	       pvalue < b.pvalue );
    }
};
 
///////////////////////////////////////////////
// Helper function to find n of m combinations

void combinations_recursive(const vector<int> &elems,
			    unsigned long req_len,
			    vector<unsigned long> &pos,
			    unsigned long depth,
			    unsigned long margin,
			    vector<vector<int> > & collection)
{
  
  if (depth >= req_len) {
    vector<int> t;
    for (unsigned long ii = 0; ii < pos.size(); ++ii)
      t.push_back( elems[pos[ii]] );
    collection.push_back(t);
    return;
  }

  if ((elems.size() - margin) < (req_len - depth))
    return;

  for (unsigned long ii = margin; ii < elems.size(); ++ii) {
    pos[depth] = ii;
    combinations_recursive(elems, req_len, pos, depth + 1, ii + 1, collection);
  }
  return;
}


///////////////////////////////////////////////
// For locus l, main proxy association function

void Plink::performProxyTests(int l)
{

  if (!par::SNP_major) 
    Ind2SNP();
  
  // Consider a particular SNP, l, and 
  // a) form haplotypes surrounding SNPs
  // b) calculate association with phenotype for these SNPs
  // c) calculate r^2 (haplotypic) with allele and haplotypes
  
  // Form haplotypes based on the surrounding SNPs
  // Form phenotype based on the patterns of missingness for test SNP
  // Is there any association?

  bool old_silent = par::silent;


  ///////////////////////
  // If we are in 'impute' mode: this is to evaluate imputation, 
  // in a 'leave-one-out' manner; here we assume the reference panel
  // is coded as 'missing phenotype' and the rest of the sample is coded 
  // as non-missing phenotype; so we must first blank out (but later replace) 
  // any genotype data for these individuals

  vector<bool> tmp1, tmp2;

  if ( par::proxy_impute || par::proxy_leave_out ) 
    {
      for ( int i = 0 ; i < n ; i++ ) 
	if ( ! sample[i]->missing )
	  {
	    tmp1.push_back( SNP[l]->one[i] );
	    tmp2.push_back( SNP[l]->two[i] );
	    
	    SNP[l]->one[i] = true;
	    SNP[l]->two[i] = false;
	  }
    }
 
  ///////////////////////
  // Form test haplotypes

  haplo->reset();
  haplo->new_pred_locus.resize(1);
  haplo->new_map.resize(1);
  
  CSNP * s = SNP[l];
  
  vector<int> proxyHaplotypePlusSNP;


  // Add reference SNP

  proxyHaplotypePlusSNP.push_back(l);

  // Either a fixed maximum number of SNPs left and right (allowing
  // for different filters, and chromosome ends) or read from a file
  // (these SNPs must be on same chromosome)
  
  if ( par::proxy_list )
    {
      checkFileExists( par::proxy_list_file );
      printLOG("Reading proxy list from [ " 
	       + par::proxy_list_file + " ]\n");
      ifstream PL( par::proxy_list_file.c_str(), ios::in);
      
      map<string,int> mlocus;
      for (int j=0;j<nl_all;j++)
	mlocus.insert(make_pair(locus[j]->name,j));

      while ( ! PL.eof() )
	{
	  string psnp;
	  PL >> psnp;

	  if ( psnp == "" )
	    continue;
	  
	  map<string,int>::iterator m = mlocus.find( psnp );
	  
	  if ( m == mlocus.end() )
	    continue;
	  
	  // Add if okay w/ MAF and genotyping thresholds
	  
	  int pn = m->second;
	  
	  if ( pn != l && 
	       locus[pn]->chr == locus[l]->chr && 
	       locus[pn]->freq >= par::proxy_maf && 
	       abs(double((locus[l]->bp - locus[pn]->bp)/1000.0)) <= par::proxy_kb &&
	       genotypingRate(*this, pn) <= par::proxy_geno )
	    {
	      proxyHaplotypePlusSNP.push_back( pn );	      
	    }
	  
	}

      PL.close();
    }
  
  else // ... use window approach
    {
      
      int i = l-1;
      int added = 0;
      
      while ( added < par::proxy_window )
	{
	  if ( i >= 0 && 
	       locus[i]->chr == locus[l]->chr )
	    {
	      
	      // Add MAF and genotyping thresholds here
	      if ( locus[i]->freq >= par::proxy_maf && 
		   abs(double((locus[l]->bp - locus[i]->bp)/1000.0)) <= par::proxy_kb &&
		   genotypingRate(*this, i) <= par::proxy_geno )
		{
		  proxyHaplotypePlusSNP.push_back(i);
		  added++;
		}
	      
	      // Shift left
	      --i;
	    }
	  else
	    {
	      // Cannot add any more
	      added = par::proxy_window;
	    }
	}
      
      // Now move right
      added = 0;
      i = l+1;
      while ( added < par::proxy_window )
	{
	  if ( i < nl_all && 
	       locus[i]->chr == locus[l]->chr )
	    {
	      
	      // Add MAF, kb and genotyping thresholds here
	      if ( locus[i]->freq >= par::proxy_maf && 
		   abs(double((locus[l]->bp - locus[i]->bp)/1000.0)) <= par::proxy_kb &&
		   genotypingRate(*this, i) <= par::proxy_geno )
		{
		  proxyHaplotypePlusSNP.push_back(i);
		  added++;
		}
	      
	      //Shift right
	      ++i;
	    }
	  else
	    {
	      // Cannot add any more
	      added = par::proxy_window;
	    }
	  
	}
    }
  
  

  // Sort, and select reference SNP

  sort( proxyHaplotypePlusSNP.begin(), 
	proxyHaplotypePlusSNP.end());

  int cnt = proxyHaplotypePlusSNP.size();
  int ref;

  for (int i=0; i<cnt; i++)
    if ( proxyHaplotypePlusSNP[i] == l )
      {
	ref = i;
	break;
      }


  haplo->new_pred_locus[0] = proxyHaplotypePlusSNP;
  haplo->new_map[0] = locus[l];
  

  ///////////////////
  // Phase haplotypes
  
  par::silent = true;
  haplo->new_pred_allele = listPossibleHaplotypes(*this,
						  haplo->new_pred_locus[0]);

  haplo->phaseAllHaplotypes();
  
  haplo->hname = locus[l]->name;
  
  par::silent = old_silent;



  /////////////////////////////////////////////////////////////////////
  // If we are in 'impute' mode: do not perform association tests, but
  // just compare imputed genotypes to actual (left-out) genotypes,
  // and report on this.

  if ( par::proxy_leave_out )
    {
      int cnt = 0;
      for ( int i = 0 ; i < n ; i++ ) 
	if ( ! sample[i]->missing )
	  {
	    SNP[l]->one[i] = tmp1[cnt];
	    SNP[l]->two[i] = tmp2[cnt++];
	  }

      tmp1.clear();
      tmp2.clear();
    }
  else if ( par::proxy_impute ) 
    {
      
      int con[4][4];
      for (int j=0;j<4;j++)
	for (int k=0; k<4; k++)
	  con[j][k] = 0; 
            
      // Replace genotypes
      int cnt = 0;
      for ( int i = 0 ; i < n ; i++ ) 
	if ( ! sample[i]->missing )
	  {
	    // Actual genotypes
	    bool a1 = tmp1[cnt];
	    bool a2 = tmp2[cnt++];

	    int og;

	    if ( a1 )
	      {
		if ( a2 ) 
		  og = 2;
		else
		  og = 3;
	      }
	    else
	      {
		if ( a2 ) 
		  og = 1;
		else
		  og = 0;
	      }
	    
	    // Imputed genotypes

	    vector_t g = haplo->imputeGenotype(i,ref);

	    // Call imputed
	    bool i1, i2;
	    double threshold = 0.8;
	    if ( g[0] > threshold ) 
	      {
		i1 = i2 = false;
		con[og][0]++;
	      }
	    else if ( g[1] > threshold ) 
	      { 
		i1 = false; 
		i2 = true; 
		con[og][1]++;
	      }
	    else if ( g[2] > threshold ) 
	      {
		i1 = i2 = true;
		con[og][2]++;
	      }
	    else 
	      { 
		i1 = true; i2 = false; 
		con[og][3]++;
	      }

	    if ( par::proxy_full_report ) 
	      haplo->HTEST << locus[l]->name << "\t" 
			   << sample[i]->fid << " " 
			   << sample[i]->iid << "\t" 
			   << a1<<a2 << " " 
			   << i1<<i2 << " " 
			   << g[0] << " " 
			   << g[1] << " " 
			   << g[2] << "\n";
	    
	    
	    // Replace originals
	    SNP[l]->one[i] = a1;
	    SNP[l]->two[i] = a2;
	  }

      // Report matrix of concordance

      int total = 0;
      for (int j=0;j<4;j++)
	for (int k=0;k<4;k++)
	  total += con[j][k];

      int concordant = con[0][0] + con[1][1] + con[2][2];

      int both_geno = 0;
      for (int j=0;j<3;j++)
	for (int k=0;k<3;k++)
	  both_geno += con[j][k];
      
      int observed_geno = total;
      for (int j=0; j<4; j++)
	observed_geno -= con[3][j];
      
      int imputed_geno = total;
      for (int j=0; j<4; j++)
	imputed_geno -= con[j][3];
      
      double rate = both_geno == 0 ? 
	0 : 
	(double)concordant/(double)both_geno ;
            
      if ( par::proxy_full_report ) 
	{
	  haplo->HTEST << "\nImputation matrix (rows observed, columns imputed)\n\n";

	  for (int j=0;j<4;j++)
	    {
	      
	      if ( j == 0 ) 
		haplo->HTEST << locus[l]->allele1 << "/"
			     << locus[l]->allele1 << "\t";
	      else if ( j == 1 ) 
		haplo->HTEST << locus[l]->allele1 << "/"
			     << locus[l]->allele2 << "\t";
	      else if ( j == 2 ) 
		haplo->HTEST << locus[l]->allele2 << "/"
			     << locus[l]->allele2 << "\t";
	      else
		haplo->HTEST << par::missing_genotype << "/"
			     << par::missing_genotype << "\t";	  
	      
	      for (int k=0; k<4; k++)
		haplo->HTEST << con[j][k] << "\t";
	      
	      haplo->HTEST << "\n";
	      

	    }
	  
	  haplo->HTEST << setw(par::pp_maxsnp) << "SNP" << " "
		       << setw(8) << "TOT_N" << " "
		       << setw(8) << "CONC" << " "
		       << setw(8) << "RATE" << " "
		       << setw(8) << "BOTH" << " "
		       << setw(8) << "OBS_N" << " "
		       << setw(8) << "IMP_N" << "\n";
	}

      haplo->HTEST << setw(par::pp_maxsnp) << locus[l]->name << " "
		   << setw(8) << total << " "
		   << setw(8) << concordant << " "
		   << setw(8) << rate << " "
		   << setw(8) << both_geno << " "
		   << setw(8) << observed_geno << " "
		   << setw(8) << imputed_geno << "\n";
      
      tmp1.clear();
      tmp2.clear();
      
      return;
    }


  
  if ( ! ( par::proxy_all || par::proxy_full_report ) )
    printLOG("Estimated haplotype frequencies: now considering combinations...\n");


  ////////////////////////////////////////////////////
  // Display haplotype frequencies and individual r^2

  if ( ( ! par::proxy_all ) || par::proxy_full_report )
    haplo->HTEST << "\n"
		 << "     *** Proxy haplotype association report for " 
		 << haplo->hname << " *** \n\n";
  

  ////////////////////////////////////////
  // Report SNPs and single-SNP r-squared
	    
  if ( ( ! par::proxy_all) || par::proxy_full_report )
    {
      haplo->HTEST << setw(par::pp_maxsnp) << "SNP" << " " 
		   << setw(8) << "MAF" << " " 
		   << setw(8) << "GENO" << " "
		   << setw(8) << "KB" << " "
		   << setw(8) << "RSQ" << " ";
      if ( par::qt ) 
	haplo->HTEST << setw(8) << "BETA" << " "
		     << setw(8) << "STAT" << " "	  
		     << setw(8) << "P" << "\n";
      else 
	haplo->HTEST << setw(8) << "OR" << " "
		     << setw(8) << "CHISQ" << " "	  
		     << setw(8) << "P" << "\n";


      for ( int i = 0 ; i < cnt ; i++)
	{	  
	  haplo->HTEST << setw(par::pp_maxsnp) 
		       << locus[ haplo->new_pred_locus[0][i] ]->name << " "
		       << setw(8) 
		       << locus[ haplo->new_pred_locus[0][i] ]->freq << " "
		       << setw(8) 
		       << genotypingRate(*this,haplo->new_pred_locus[0][i]) << " "
		       << setw(8) 
		       << (double)(locus[ haplo->new_pred_locus[0][i]]->bp - 
				   locus[ haplo->new_pred_locus[0][ref]]->bp)/1000<<" ";
	  
	  // R-squared

	  if ( i == ref ) 
	    haplo->HTEST << setw(8) << "*" << " ";
	  else
	    haplo->HTEST << setw(8) << haplo->rsq_internal(i,ref) << " ";

	  // Single SNP association

	  boolvec_t snpmask(cnt,false);
	  boolvec_t dummy_allele(cnt,true);
	  snpmask[i] = true;
	  
	  map<int,int> tests = haplo->makeTestSet(snpmask,dummy_allele);
	  
	  if (par::proxy_CC)
	    {
	      if ( par::qt )
		haplo->haplotypicQTL(tests,2,false);
	      else
		haplo->haplotypicCC(tests,2,false);
	    }
	  else if (par::proxy_TDT)
	    haplo->haplotypicTDT(tests,2,false);
	  
	  haplo->HTEST << setw(8) << haplo->odds << " "
		       << setw(8) << haplo->result << " "
		       << setw(8) << haplo->pvalue << "\n";
	  
	  
	}      
      
      haplo->HTEST << "\n\n";
    }

  
  //////////////////////////////////////
  // Report haplotypes and frequencies, and also 
  // all haplotype-specific tests and omnibus test 
  // result
  
  if ( ( ! par::proxy_all ) || par::proxy_full_report )
    {
      string str = "";
      for ( int i = 0; i < cnt ; i++)
	if ( i == ref ) 
	  str += "*";
	else
	  str += ".";
      
      haplo->HTEST << setw(14) << str << " "
		   << setw(10) << "FREQ" << " ";
      if ( par::qt ) 
	haplo->HTEST << setw(8) << "BETA" << " "
		     << setw(8) << "STAT" << " "	  
		     << setw(8) << "P" << "\n";
      else 
	haplo->HTEST << setw(10) << "OR" << " "
		     << setw(10) << "CHISQ" << " " 
		     << setw(10) << "P" << "\n";

      for (int h=0; h< haplo->nh; h++)      
	{
	  if (haplo->f[h] >= par::proxy_mhf )
	    {	  
	      haplo->HTEST << setw(14) << haplo->haplotypeName(h) << " " 
			   << setw(10) << haplo->f[h] << " ";
	      
	      map<int,int> tests;
	      
	      for (int h2=0; h2 < haplo->nh; h2++)
		{
		  if ( haplo->f[h2] >= par::proxy_mhf)
		    {
		      if (h==h2) 
			{
			  tests.insert(make_pair(h2,0));
			}
		      else 
			{
			  tests.insert(make_pair(h2,1));
			}
		    }
		}

	      if (par::proxy_CC)      
		{
		  if ( par::qt)
		    haplo->haplotypicQTL(tests,2,false);		
		  else
		    haplo->haplotypicCC(tests,2,false);		
		}
	      else if (par::proxy_TDT)
		haplo->haplotypicTDT(tests,2,false);
	      
	      haplo->HTEST << setw(10) << haplo->odds << " "
			   << setw(10) << haplo->result << " "
			   << setw(10) << haplo->pvalue << "\n";
	      
	    }
	}   


      haplo->HTEST << "\nHaplotype frequency estimation based on " 
		   << haplo->validN; 

      if ( haplo->X ) 
	{
	  int found_chr = 0;
	  for (int i=0; i<n; i++)
	    if ( sample[i]->founder )
	      {
		if ( sample[i]->sex )
		  found_chr++;
		else
		  found_chr+=2;
	      }
	  haplo->HTEST << " of " << found_chr << " founder chromosomes\n";
	}
      else if ( haplo->haploid )
	haplo->HTEST << " of " << haplo->cnt_f << " founder chromosomes\n";
      else
	haplo->HTEST << " of " << haplo->cnt_f * 2 << " founder chromosomes\n";

		   

      
      // Omnibus test: C/C only

      if ( par::proxy_CC && ! par::qt )
	{
	   
	  map<int,int> tests;
	  int nch=0;
	  for (int h=0; h < haplo->nh; h++)
	    if ( haplo->f[h] >= par::proxy_mhf)
	      tests.insert(make_pair(h,nch++));
      
	  if (nch>2)
	    {
	      if (par::proxy_CC)
		haplo->haplotypicCC(tests,nch,false);
	      else if (par::proxy_TDT)
		haplo->haplotypicTDT(tests,nch,false);
	      
	      haplo->HTEST << "Omnibus haplotype test statistic: " 
			   << haplo->result << ", df = " << nch-1 << ", "
			   << "p = " 
			   << chiprobP( haplo->result , nch-1 ) << "\n\n";
	    }
	  
	} 
    }


  ///////////////////////////////////////////
  // Create masks

  // Reference SNP (fix here)
  boolvec_t m1(cnt,false);
  m1[ref] = true;
  boolvec_t a1(cnt,false);

  // Proxy haplotype (populated below)
  boolvec_t m2(cnt,false);
  boolvec_t a2(cnt); 
  
  
  ///////////////////////////////////////////
  // Report actual SNP

  // Note: for the TDT, the nested haplotype test has some different
  // properties (i.e. SNP that is homozygous might fall on
  // heterozygous haplotypes, and therefore be counted. At the
  // extreme, this will lead to a test more like the HHRR.  In any 
  // case, apply this model twice to each allele separately, as these
  // are not equivalent -- might be interesting to look into this, but 
  // probably just restrict to the C/C scenario.  Otherwise we need to 
  // re-score informativeness each time a new. 

  map<int,int> tests = haplo->makeTestSet(m1,a1);

  if (par::proxy_CC)
    {
      if ( par::qt )
	haplo->haplotypicQTL(tests,2,false);
      else
	haplo->haplotypicCC(tests,2,false);
    }
  else if (par::proxy_TDT)
    haplo->haplotypicTDT(tests,2,false);

  double stored_result = haplo->result;
  double stored_pvalue = haplo->pvalue;

 
  // REMOVE THIS SECTION WHEN PROPER TDT IN PLACE
  // Perform second experimental TDT test
  //   if (par::proxy_TDT)
  //     {
  
  //       map<int,int> tests;  
  //       for (int h2=0; h2 < haplo->nh; h2++)
  // 	{
  // 	  bool is_A = true;
  
  // 	  for (int s = 0; s < haplo->ns ; s++)
  // 	    if ( m1[s] && haplo->hap[h2][s] != a1[s] )
  // 	      is_A = false;
  
  // 	  if ( is_A ) 
  // 	    tests.insert(make_pair(h2,0));
  // 	  else 
  // 	    tests.insert(make_pair(h2,1));      
  // 	}
  
  //       haplo->haplotypicTDT(tests,2,false);
  //     }
  // REMOVE THIS SECTION WHEN PROPER TDT IN PLACE



  // Just report the single SNP result (based on haplotype test)?

  if ( par::proxy_all && ( ! par::proxy_full_report ) )
    {
      haplo->HTEST << setw( 4 ) << locus[l]->chr << " " 
		   << setw(par::pp_maxsnp) << locus[l]->name << " "
		   << setw(8) << stored_pvalue;
      
      // REMOVE WHEN PROPER TDT IN PLACE
      // if ( par::proxy_TDT )
      //    haplo->HTEST << setw(8) << haplo->pvalue;
      
      haplo->HTEST << "\n";
      return;
    }

  
  /////////////////////////////////////////////////////////
  // Rest of this function is for the extended report mode
  
    
    
  ///////////////////////////////////////////
  // Consider from cnt-1 to 1 SNP haplotypes

  int search_cnt = par::proxy_include_reference ? cnt : cnt - 1;
  int maxsnp = search_cnt < par::proxy_maxhap ? search_cnt : par::proxy_maxhap;
  
  set<ProxyResult> presults;
  
  for (int i=1; i<=maxsnp; i++)
    {

      // For an i-SNP of cnt-SNP, consider all the permutations

      vector<unsigned long> pos1(i);
      vector<int> d(search_cnt);
      int i2=0;
      for (int z=0; z<cnt; z++) 
	if ( z != ref ) 
	  {
	    d[i2]=i2;
	    i2++;
	  }

      vector<vector<int> > collection;
      combinations_recursive(d,i,pos1,0,0,collection);

      vector<int> mapback;
      for (int s=0;s<cnt;s++)
	if ( par::proxy_include_reference || s != ref )
	  mapback.push_back(s);
      
      // Consider each combination of i of cnt-1 SNP haplotypes
      // then consider each possible haplotype within that
      
      for (int c1 = 0; c1 < collection.size() ; c1++ )
	{
	  
	  // For biallelic haplotypes (SNPs)
	  // skip the second allele
	  
	  // Which SNPs do we want to look up?
	  
	  vector<int> posit = collection[c1];
	  
	  // Now consider all search_cnt SNP haplotypes 
	  // i.e. excluding reference SNP
	  
	  int hapcnt = (int)pow((double)2,i);
	  int h = 0;
	  while ( h < hapcnt )
	    {
	      
	      // Skip redundant second allele of 
	      // SNPs

	      if ( collection[c1].size() == 1 && 
		   h == 1 )
		{
		  h++;
		  continue;
		}
	      
	      
	      vector<bool> tmp;
	      unsigned int p=1;
	      for (int s=0;s<i;s++)
		{
		  if ( h & p ) tmp.push_back(false);
		  else tmp.push_back(true);
		  p <<= 1;
		}
	      
	      
	      ///////////////////////
	      // Set proxy haplotype
	      
	      // Insert i-SNP haplotype back into 
	      // full cnt-SNP space
	      
	      fill(m2.begin(), m2.end(), false);
	      
	      int t1=0;
	      for (int t2=0; t2<i; t2++)
		{
		  m2[ mapback[ posit[t2] ] ] = true;
		  a2[ mapback[ posit[t2] ] ] = tmp[ t2 ];
		}
	      
	      
	      ////////////////////////////////
	      // Check frequency of haplotype
	      
	      double f = haplo->freq(m2,a2);
	      
	      
	      ////////////////////////////////////////////////////////////
	      // Calculate r^2 between haplotype and reference SNP allele
	      
	      double r2 = haplo->rsq_internal(m1,a1,m2,a2);
	      

	      ////////////////////////////////////////////
	      // Is this haplotype not worth considering?
	     
 	      if ( r2 < par::proxy_r2 ||
		   f < par::proxy_mhf )
 		{		  
 		  h++;
 		  continue;
		}

	      ////////////////////////////////////////////////////////////
	      // Calculte association between proxy haplotype and disease
	      
	      // If only two haplotypes, report only 1
	      // (note may only be two *common* haplotypes, but
	      // in that case, we should report both
	      
	      map<int,int> tests = haplo->makeTestSet(m2,a2);
		      
	      if (par::proxy_CC)
		{
		  if ( par::qt )
		    haplo->haplotypicQTL(tests,2,false);
		  else
		    haplo->haplotypicCC(tests,2,false);
		}
	      else if (par::proxy_TDT)
 		{
		  haplo->haplotypicTDT(tests,2,false);
		}

	      string str = par::proxy_include_reference ? 
		haplo->getSubHaplotypeName(m2,a2,-1) :
		haplo->getSubHaplotypeName(m2,a2,ref);
	      
	      //////////////////////
	      // Store this result


	      ProxyResult r(str,f,r2,
			    haplo->odds,
			    haplo->result,
			    haplo->pvalue);
	      presults.insert(r);
	      
	      // Consider next haplotype
	      
	      h++;
	      
	    }
	}      
    }


  // Report results

  if ( presults.size() == 0 )
    haplo->HTEST << "No proxies found above r-sq " << par::proxy_r2 << "\n";
  else
    {
      
      haplo->HTEST << setw(14) << "HAP" << " "
		   << setw(10) << "FREQ" << " " 
		   << setw(10) << "RSQ" << " ";
      if ( par::qt ) 
	haplo->HTEST << setw(8) << "BETA" << " "
		     << setw(8) << "STAT" << " "	  
		     << setw(8) << "P" << "\n";
      else 
	haplo->HTEST << setw(10) << "OR" << " "
		     << setw(10) << "CHISQ" << " "
		     << setw(10) << "P" << "\n";
      
      
      set<ProxyResult>::iterator i = presults.begin();
      
      while ( i != presults.end() )
	{
	  haplo->HTEST << setw(14) << i->name << " "
		       << setw(10) << i->f << " " 
		       << setw(10) << i->r2 << " " 
		       << setw(10) << i->odds << " " 
		       << setw(10) << i->chisq << " " 
		       << setw(10) << i->pvalue << "\n";
	  i++;
	}
    }

  if ( par::proxy_full_report ) 
    haplo->HTEST << "\n+--------------------------------------------------------------------+\n\n";

  return;
  
  
}


